(*
  LibOVR wrapper for Delphi

  Copyright (c) 2014 Lyna

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
*)
unit LibOVR;

interface

{$MINENUMSIZE 4}

const
  OVR_MAJOR_VERSION = 0;
  OVR_MINOR_VERSION = 4;
  OVR_BUILD_VERSION = 2;
  OVR_VERSION_STRING = '0.4.2';

type
{$IFNDEF NativeUInt}
  NativeUInt = Cardinal;
{$ENDIF}

  ovrBool = Boolean;

//{$DEFINE ENABLE_LATENCY_TESTER}

//------------------------------------------------------------------------------
// ***** Simple Math Structures

/// A 2D vector with integer components.
  ovrVector2i = record
    x, y: Longint;
  end;
/// A 2D size with integer components.
  ovrSizei = record
    w, h: Longint;
  end;
  PovrSizei = ^ovrSizei;
/// A 2D rectangle with a position and size.
/// All components are integers.
  ovrRecti = record
    Pos: ovrVector2i;
    Size: ovrSizei;
  end;
  PovrRecti = ^ovrRecti;

/// A quaternion rotation.
  ovrQuatf = record
    x, y, z, w: Single;
  end;
/// A 2D vector with float components.
  ovrVector2f = record
    x, y: Single;
  end;
/// A 3D vector with float components.
  ovrVector3f = record
    x, y, z: Single;
  end;
/// A 4x4 matrix with float elements.
  ovrMatrix4f = record
    M: array[0..3, 0..3] of Single;
  end;
/// Position and orientation together.
  ovrPosef = record
    Orientation: ovrQuatf;
    Position: ovrVector3f;
  end;

/// A full pose (rigid body) configuration with first and second derivatives.
  ovrPoseStatef = record
    ThePose: ovrPosef;
    AngularVelocity: ovrVector3f;
    LinearVelocity: ovrVector3f;
    AngularAcceleration: ovrVector3f;
    LinearAcceleration: ovrVector3f;
    TimeInSeconds: Double;         // Absolute time of this state sample.
  end;

/// Field Of View (FOV) in tangent of the angle units.
/// As an example, for a standard 90 degree vertical FOV, we would
/// have: { UpTan = tan(90 degrees / 2), DownTan = tan(90 degrees / 2) }.
  ovrFovPort = record
    /// The tangent of the angle between the viewing vector and the top edge of the field of view.
    UpTan: Single;
    /// The tangent of the angle between the viewing vector and the bottom edge of the field of view.
    DownTan: Single;
    /// The tangent of the angle between the viewing vector and the left edge of the field of view.
    LeftTan: Single;
    /// The tangent of the angle between the viewing vector and the right edge of the field of view.
    RightTan: Single;
  end;


//------------------------------------------------------------------------------
// ***** HMD Types

/// Enumerates all HMD types that we support.
  ovrHmdType = (
    ovrHmd_None  = 0,
    ovrHmd_DK1   = 3,
    ovrHmd_DKHD  = 4,
    ovrHmd_DK2   = 6,
    ovrHmd_Other  // Some HMD other then the one in the enumeration.
  );

/// HMD capability bits reported by device.      windows
  ovrHmdCaps = Longword;

const
  // Read-only flags.
  ovrHmdCap_Present           = $0001;   ///  The HMD is plugged in and detected by the system.
  ovrHmdCap_Available         = $0002;   ///  The HMD and its sensor is available for ownership use.
                                         ///  i.e. it is not already owned by another application.
  ovrHmdCap_Captured          = $0004;   ///  Set to 'true' if we captured ownership of this HMD.

  // These flags are intended for use with the new driver display mode.
  ovrHmdCap_ExtendDesktop     = $0008;   /// (read only) Means the display driver is in compatibility mode.

  // Modifiable flags (through ovrHmd_SetEnabledCaps).
  ovrHmdCap_NoMirrorToWindow  = $2000;   /// Disables mirroring of HMD output to the window. This may improve
                                         /// rendering performance slightly (only if 'ExtendDesktop' is off).
  ovrHmdCap_DisplayOff        = $0040;   /// Turns off HMD screen and output (only if 'ExtendDesktop' is off).

  ovrHmdCap_LowPersistence    = $0080;   ///  HMD supports low persistence mode.
  ovrHmdCap_DynamicPrediction = $0200;   ///  Adjust prediction dynamically based on internally measured latency.
  /// Support rendering without VSync for debugging.
  ovrHmdCap_NoVSync           = $1000;

  // These bits can be modified by ovrHmd_SetEnabledCaps.
  ovrHmdCap_Writable_Mask     = $33F0;
  /// These flags are currently passed into the service. May change without notice.
  ovrHmdCap_Service_Mask      = $23F0;


/// Tracking capability bits reported by the device.
/// Used with ovrHmd_ConfigureTracking.
type
  ovrTrackingCaps = Longword;

const
  ovrTrackingCap_Orientation      = $0010;   ///  Supports orientation tracking (IMU).
  ovrTrackingCap_MagYawCorrection = $0020;   ///  Supports yaw drift correction via a magnetometer or other means.
  ovrTrackingCap_Position         = $0040;   ///  Supports positional tracking.
  ///  Overrides the other flags. Indicates that the application
  ///  doesn't care about tracking settings. This is the internal
  ///  default before ovrHmd_ConfigureTracking is called.
  ovrTrackingCap_Idle             = $0100;

/// Distortion capability bits reported by device.
/// Used with ovrHmd_ConfigureRendering and ovrHmd_CreateDistortionMesh.
type
  ovrDistortionCaps = Longword;

const
  ovrDistortionCap_Chromatic      = $01;  ///  Supports chromatic aberration correction.
  ovrDistortionCap_TimeWarp       = $02;  ///  Supports timewarp.
  ovrDistortionCap_Vignette       = $08;  ///  Supports vignetting around the edges of the view.
  ovrDistortionCap_NoRestore      = $10;  ///  Do not save and restore the graphics state when rendering distortion.
  ovrDistortionCap_FlipInput      = $20;  ///  Flip the vertical texture coordinate of input images.
  ovrDistortionCap_SRGB           = $40;  ///  Assume input images are in sRGB gamma-corrected color space.
  ovrDistortionCap_Overdrive      = $80;  ///  Overdrive brightness transitions to reduce artifacts on DK2+ displays
  ovrDistortionCap_HqDistortion   = $100; ///  High-quality sampling of distortion buffer for anti-aliasing

  ovrDistortionCap_ProfileNoTimewarpSpinWaits = $10000;  /// Use when profiling with timewarp to remove false positives


/// Specifies which eye is being used for rendering.
/// This type explicitly does not include a third "NoStereo" option, as such is
/// not required for an HMD-centered API.
type
  ovrEyeType = (
    ovrEye_Left  = 0,
    ovrEye_Right = 1,
    ovrEye_Count = 2
  );



/// This is a complete descriptor of the HMD.
  ovrHmdDesc = record
    /// Internal handle of this HMD.
    Handle: Pointer; //struct ovrHmdStruct*

    /// This HMD's type.
    Type_: ovrHmdType;

    /// Name string describing the product: "Oculus Rift DK1", etc.
    ProductName: PAnsiChar;
    Manufacturer: PAnsiChar;

    /// HID Vendor and ProductId of the device.
    VendorId: SmallInt;
    ProductId: SmallInt;
    /// Sensor (and display) serial number.
    SerialNumber: array[0..23] of AnsiChar;
    /// Sensor firmware version.
    FirmwareMajor: SmallInt;
    FirmwareMinor: SmallInt;
    /// External tracking camera frustum dimensions (if present).
    CameraFrustumHFovInRadians: Single;
    CameraFrustumVFovInRadians: Single;
    CameraFrustumNearZInMeters: Single;
    CameraFrustumFarZInMeters: Single;

    /// Capability bits described by ovrHmdCaps.
    HmdCaps: Longword;
    /// Capability bits described by ovrTrackingCaps.
    TrackingCaps: Longword;
    /// Capability bits described by ovrDistortionCaps.
    DistortionCaps: Longword;

    /// These define the recommended and maximum optical FOVs for the HMD.
    DefaultEyeFov: array[0..Ord(ovrEye_Count)-1] of ovrFovPort;
    MaxEyeFov: array[0..Ord(ovrEye_Count)-1] of ovrFovPort;

    /// Preferred eye rendering order for best performance.
    /// Can help reduce latency on sideways-scanned screens.
    EyeRenderOrder: array[0..Ord(ovrEye_Count)-1] of ovrEyeType;

    /// Resolution of the full HMD screen (both eyes) in pixels.
    Resolution: ovrSizei;
    /// Location of the application window on the desktop (or 0,0).
    WindowsPos: ovrVector2i;

    /// Display that the HMD should present on.
    /// TBD: It may be good to remove this information relying on WindowPos instead.
    /// Ultimately, we may need to come up with a more convenient alternative,
    /// such as API-specific functions that return adapter, or something that will
    /// work with our monitor driver.
    /// Windows: (e.g. "\\\\.\\DISPLAY3", can be used in EnumDisplaySettings/CreateDC).
    DisplayDeviceName: PAnsiChar;
    /// MacOS:
    DisplayId: Longint;
  end;


/// Simple type ovrHmd is used in ovrHmd_* calls.
  ovrHmd = ^ovrHmdDesc;



/// Bit flags describing the current status of sensor tracking.
  ovrStatusBits = Longword;

const
  ovrStatus_OrientationTracked = $0001;   /// Orientation is currently tracked (connected and in use).
  ovrStatus_PositionTracked    = $0002;   /// Position is currently tracked (false if out of range).
  ovrStatus_CameraPoseTracked  = $0004;   /// Camera pose is currently tracked.
  ovrStatus_PositionConnected  = $0020;   /// Position tracking hardware is connected.
  ovrStatus_HmdConnected       = $0080;   /// HMD Display is available and connected.

/// Specifies a reading we can query from the sensor.
type
  ovrSensorData = record
    Accelerometer: ovrVector3f; /// Acceleration reading in m/s^2.
    Gyro: ovrVector3f;          /// Rotation rate in rad/s.
    Magnetometer: ovrVector3f;  /// Magnetic field in Gauss.
    Temperature: Single;        /// Temperature of the sensor in degrees Celsius.
    TimeInSeconds: Single;      /// Time when the reported IMU reading took place, in seconds.
  end;


/// Tracking state at a given absolute time (describes predicted HMD pose etc).
/// Returned by ovrHmd_GetTrackingState.
  ovrTrackingState = record
    /// Predicted head pose (and derivatives) at the requested absolute time.
    /// The look-ahead interval is equal to (HeadPose.TimeInSeconds - RawSensorData.TimeInSeconds).
    HeadPose: ovrPoseStatef;

    /// Current pose of the external camera (if present).
    /// This pose includes camera tilt (roll and pitch). For a leveled coordinate
    /// system use LeveledCameraPose.
    CameraPose: ovrPosef;

    /// Camera frame aligned with gravity.
    /// This value includes position and yaw of the camera, but not roll and pitch.
    /// It can be used as a reference point to render real-world objects in the correct location.
    LeveledCameraPose: ovrPosef;

    /// The most recent sensor data received from the HMD.
    RawSensorData: ovrSensorData;

    /// Tracking status described by ovrStatusBits.
    StatusFlags: Longword;

    //// 0.4.1

    /// Time spent processing the last vision frame
    LastVisionProcessingTime: Double;
  end;


/// Frame timing data reported by ovrHmd_BeginFrameTiming() or ovrHmd_BeginFrame().
  ovrFrameTiming = record
    /// The amount of time that has passed since the previous frame's
    /// ThisFrameSeconds value (usable for movement scaling).
    /// This will be clamped to no more than 0.1 seconds to prevent
    /// excessive movement after pauses due to loading or initialization.
    DeltaSeconds: Single;

    /// It is generally expected that the following holds:
    /// ThisFrameSeconds < TimewarpPointSeconds < NextFrameSeconds <
    /// EyeScanoutSeconds[EyeOrder[0]] <= ScanoutMidpointSeconds <= EyeScanoutSeconds[EyeOrder[1]].

    /// Absolute time value when rendering of this frame began or is expected to
    /// begin. Generally equal to NextFrameSeconds of the previous frame. Can be used
    /// for animation timing.
    ThisFrameSeconds: Double;
    /// Absolute point when IMU expects to be sampled for this frame.
    TimewarpPointSeconds: Double;
    /// Absolute time when frame Present followed by GPU Flush will finish and the next frame begins.
    NextFrameSeconds: Double;

    /// Time when when half of the screen will be scanned out. Can be passed as an absolute time
    /// to ovrHmd_GetTrackingState() to get the predicted general orientation.
    ScanoutMidpointSeconds: Double;
    /// Timing points when each eye will be scanned out to display. Used when rendering each eye.
    EyeScanoutSeconds: array[0..1] of Double;
  end;



/// Rendering information for each eye. Computed by either ovrHmd_ConfigureRendering()
/// or ovrHmd_GetRenderDesc() based on the specified FOV. Note that the rendering viewport
/// is not included here as it can be specified separately and modified per frame through:
///    (a) ovrHmd_GetRenderScaleAndOffset in the case of client rendered distortion,
/// or (b) passing different values via ovrTexture in the case of SDK rendered distortion.
  ovrEyeRenderDesc = record
    Eye: ovrEyeType;
    Fov: ovrFovPort;
    DistortedViewport: ovrRecti;            /// Distortion viewport.
    PixelsPerTanAngleAtCenter: ovrVector2f; /// How many display pixels will fit in tan(angle) = 1.
    ViewAdjust: ovrVector3f;                /// Translation to be applied to view matrix.
  end;


//------------------------------------------------------------------------------
// ***** Platform-independent Rendering Configuration

/// These types are used to hide platform-specific details when passing
/// render device, OS, and texture data to the API.
///
/// The benefit of having these wrappers versus platform-specific API functions is
/// that they allow game glue code to be portable. A typical example is an
/// engine that has multiple back ends, say GL and D3D. Portable code that calls
/// these back ends may also use LibOVR. To do this, back ends can be modified
/// to return portable types such as ovrTexture and ovrRenderAPIConfig.
  ovrRenderAPIType = (
    ovrRenderAPI_None,
    ovrRenderAPI_OpenGL,
    ovrRenderAPI_Android_GLES,  // May include extra native window pointers, etc.
    ovrRenderAPI_D3D9,
    ovrRenderAPI_D3D10,
    ovrRenderAPI_D3D11,
    ovrRenderAPI_Count
  );

/// Platform-independent part of rendering API-configuration data.
/// It is a part of ovrRenderAPIConfig, passed to ovrHmd_Configure.
  ovrRenderAPIConfigHeader = record
    API: ovrRenderAPIType;
    RTSize: ovrSizei;
    Multisample: Longint;
  end;

/// Contains platform-specific information for rendering.
  ovrRenderAPIConfig = record
    Header: ovrRenderAPIConfigHeader;
    PlatformData: array[0..7] of NativeUInt;
  end;

/// Platform-independent part of the eye texture descriptor.
/// It is a part of ovrTexture, passed to ovrHmd_EndFrame.
/// If RenderViewport is all zeros then the full texture will be used.
  ovrTextureHeader = record
    API: ovrRenderAPIType;
    TextureSize: ovrSizei;
    RenderViewport: ovrRecti;  // Pixel viewport in texture that holds eye image.
  end;

/// Contains platform-specific information about a texture.
  ovrTexture = record
    Header: ovrTextureHeader;
    PlatformData: array[0..7] of NativeUInt;
  end;


// -----------------------------------------------------------------------------
// ***** API Interfaces

// Basic steps to use the API:
//
// Setup:
//  1. ovr_Initialize()
//  2. ovrHMD hmd = ovrHmd_Create(0)
//  3. Use hmd members and ovrHmd_GetFovTextureSize() to determine graphics configuration.
//  4. Call ovrHmd_ConfigureTracking() to configure and initialize tracking.
//  5. Call ovrHmd_ConfigureRendering() to setup graphics for SDK rendering,
//     which is the preferred approach.
//     Please refer to "Client Distorton Rendering" below if you prefer to do that instead.
//  6. If the ovrHmdCap_ExtendDesktop flag is not set, then use ovrHmd_AttachToWindow to
//     associate the relevant application window with the hmd.
//  5. Allocate render target textures as needed.
//
// Game Loop:
//  6. Call ovrHmd_BeginFrame() to get the current frame timing information.
//  7. Render each eye using ovrHmd_GetEyePose to get the predicted head pose.
//  8. Call ovrHmd_EndFrame() to render the distorted textures to the back buffer
//     and present them on the hmd.
//
// Shutdown:
//  9. ovrHmd_Destroy(hmd)
//  10. ovr_Shutdown()
//

// ovr_InitializeRenderingShim initializes the rendering shim appart from everything
// else in LibOVR. This may be helpful if the application prefers to avoid
// creating any OVR resources (allocations, service connections, etc) at this point.
// ovr_InitializeRenderingShim does not bring up anything within LibOVR except the
// necessary hooks to enable the Direct-to-Rift functionality.
//
// Either ovr_InitializeRenderingShim() or ovr_Initialize() must be called before any
// Direct3D or OpenGL initilization is done by applictaion (creation of devices, etc).
// ovr_Initialize() must still be called after to use the rest of LibOVR APIs.
procedure ovr_InitializeRenderingShim; cdecl;

// Library init/shutdown, must be called around all other OVR code.
// No other functions calls besides ovr_InitializeRenderingShim are allowed
// before ovr_Initialize succeeds or after ovr_Shutdown.
/// Initializes all Oculus functionality.
function ovr_Initialize: ovrBool; cdecl;
/// Shuts down all Oculus functionality.
procedure ovr_Shutdown; cdecl;

/// Returns version string representing libOVR version. Static, so
/// string remains valid for app lifespan
function ovr_GetVersionString: PAnsiChar; cdecl;



/// Detects or re-detects HMDs and reports the total number detected.
/// Users can get information about each HMD by calling ovrHmd_Create with an index.
function ovrHmd_Detect: Longint; cdecl;


/// Creates a handle to an HMD which doubles as a description structure.
/// Index can [0 .. ovrHmd_Detect()-1]. Index mappings can cange after each ovrHmd_Detect call.
/// If not null, then the returned handle must be freed with ovrHmd_Destroy.
function ovrHmd_Create(index: Longint): ovrHmd; cdecl;
procedure ovrHmd_Destroy(hmd: ovrHmd); cdecl;

/// Creates a 'fake' HMD used for debugging only. This is not tied to specific hardware,
/// but may be used to debug some of the related rendering.
function ovrHmd_CreateDebug(type_: ovrHmdType): ovrHmd; cdecl;


/// Returns last error for HMD state. Returns null for no error.
/// String is valid until next call or GetLastError or HMD is destroyed.
/// Pass null hmd to get global errors (during create etc).
function ovrHmd_GetLastError(hmd: ovrHmd): PAnsiChar; cdecl;

/// Platform specific function to specify the application window whose output will be
/// displayed on the HMD. Only used if the ovrHmdCap_ExtendDesktop flag is false.
///   Windows: SwapChain associated with this window will be displayed on the HMD.
///            Specify 'destMirrorRect' in window coordinates to indicate an area
///            of the render target output that will be mirrored from 'sourceRenderTargetRect'.
///            Null pointers mean "full size".
/// @note Source and dest mirror rects are not yet implemented.
function ovrHmd_AttachToWindow(hmd: ovrHmd; window: THandle;
  const destMirrorRect: ovrRecti;
  const sourceRenderTargetRect: ovrRecti): ovrBool; cdecl; overload;
function ovrHmd_AttachToWindow(hmd: ovrHmd; window: THandle;
  destMirrorRect: PovrRecti = nil;
  sourceRenderTargetRect: PovrRecti = nil): ovrBool; cdecl; overload;
//------------------------------------------------------------------------------

/// Returns capability bits that are enabled at this time as described by ovrHmdCaps.
/// Note that this value is different font ovrHmdDesc::HmdCaps, which describes what
/// capabilities are available for that HMD.
function ovrHmd_GetEnabledCaps(hmd: ovrHmd): Longword; cdecl;

/// Modifies capability bits described by ovrHmdCaps that can be modified,
/// such as ovrHmd_LowPersistance.
procedure ovrHmd_SetEnabledCaps(hmd: ovrHmd; hmdCaps: Longword); cdecl;


//------------------------------------------------------------------------------
// ***** Tracking Interface

/// All tracking interface functions are thread-safe, allowing tracking state to be sampled
/// from different threads.
/// ConfigureTracking starts sensor sampling, enabling specified capabilities,
///    described by ovrTrackingCaps.
///  - supportedTrackingCaps specifies support that is requested. The function will succeed
///    even if these caps are not available (i.e. sensor or camera is unplugged). Support
///    will automatically be enabled if such device is plugged in later. Software should
///    check ovrTrackingState.StatusFlags for real-time status.
///  - requiredTrackingCaps specify sensor capabilities required at the time of the call.
///    If they are not available, the function will fail. Pass 0 if only specifying
///    supportedTrackingCaps.
///  - Pass 0 for both supportedTrackingCaps and requiredTrackingCaps to disable tracking.
function ovrHmd_ConfigureTracking(hmd: ovrHmd;
  supportedTrackingCaps, requiredTrackingCaps: Longword): ovrBool; cdecl;

/// Re-centers the sensor orientation.
/// Normally this will recenter the (x,y,z) translational components and the yaw
/// component of orientation.
procedure ovrHmd_RecenterPose(hmd: ovrHmd); cdecl;

/// Returns tracking state reading based on the specified absolute system time.
/// Pass an absTime value of 0.0 to request the most recent sensor reading. In this case
/// both PredictedPose and SamplePose will have the same value.
/// ovrHmd_GetEyePose relies on this internally.
/// This may also be used for more refined timing of FrontBuffer rendering logic, etc.
function ovrHmd_GetTrackingState(hmd: ovrHmd; absTime: Double): ovrTrackingState; cdecl;


//------------------------------------------------------------------------------
// ***** Graphics Setup


/// Calculates the recommended texture size for rendering a given eye within the HMD
/// with a given FOV cone. Higher FOV will generally require larger textures to
/// maintain quality.
///  - pixelsPerDisplayPixel specifies the ratio of the number of render target pixels
///    to display pixels at the center of distortion. 1.0 is the default value. Lower
///    values can improve performance.
function ovrHmd_GetFovTextureSize(hmd: ovrHmd; eye: ovrEyeType; fov: ovrFovPort;
  pixelsPerDisplayPixel: Single): ovrSizei;



//------------------------------------------------------------------------------
// *****  Rendering API Thread Safety

//  All of rendering functions including the configure and frame functions
// are *NOT thread safe*. It is ok to use ConfigureRendering on one thread and handle
//  frames on another thread, but explicit synchronization must be done since
//  functions that depend on configured state are not reentrant.
//
//  As an extra requirement, any of the following calls must be done on
//  the render thread, which is the same thread that calls ovrHmd_BeginFrame
//  or ovrHmd_BeginFrameTiming.
//    - ovrHmd_EndFrame
//    - ovrHmd_GetEyePose
//    - ovrHmd_GetEyeTimewarpMatrices


//------------------------------------------------------------------------------
// *****  SDK Distortion Rendering Functions

// These functions support rendering of distortion by the SDK through direct
// access to the underlying rendering API, such as D3D or GL.
// This is the recommended approach since it allows better support for future
// Oculus hardware, and enables a range of low-level optimizations.


/// Configures rendering and fills in computed render parameters.
/// This function can be called multiple times to change rendering settings.
/// eyeRenderDescOut is a pointer to an array of two ovrEyeRenderDesc structs
/// that are used to return complete rendering information for each eye.
///
///  - apiConfig provides D3D/OpenGL specific parameters. Pass null
///    to shutdown rendering and release all resources.
///  - distortionCaps describe desired distortion settings.
///
type
  TeyeFovIn = array[0..1] of ovrFovPort;
  TeyeRenderDescOut = array[0..1] of ovrEyeRenderDesc;
function ovrHmd_ConfigureRendering(hmd: ovrHmd;
  const apiConfig: ovrRenderAPIConfig; distortionCaps: Longword;
  const eyeFovIn: TeyeFovIn; const eyeRenderDescOut: TeyeRenderDescOut): ovrBool; cdecl;


/// Begins a frame, returning timing information.
/// This should be called at the beginning of the game rendering loop (on the render thread).
/// Pass 0 for the frame index if not using ovrHmd_GetFrameTiming.
function ovrHmd_BeginFrame(hmd: ovrHmd; frameIndex: Longword): ovrFrameTiming; cdecl;

/// Ends a frame, submitting the rendered textures to the frame buffer.
/// - RenderViewport within each eyeTexture can change per frame if necessary.
/// - 'renderPose' will typically be the value returned from ovrHmd_GetEyePose,
///   but can be different if a different head pose was used for rendering.
/// - This may perform distortion and scaling internally, assuming is it not
///   delegated to another thread.
/// - Must be called on the same thread as BeginFrame.
/// - *** This Function will call Present/SwapBuffers and potentially wait for GPU Sync ***.
type
  TrenderPose = array[ovrEye_Left..ovrEye_Right] of ovrPosef;
  TeyeTexture = array[0..1] of ovrTexture;
procedure ovrHmd_EndFrame(hmd: ovrHmd; const renderPose: TrenderPose;
  const eyeTexture: TeyeTexture); cdecl;


/// Returns the predicted head pose to use when rendering the specified eye.
/// - Must be called between ovrHmd_BeginFrameTiming and ovrHmd_EndFrameTiming.
/// - If the pose is used for rendering the eye, it should be passed to ovrHmd_EndFrame.
function ovrHmd_GetEyePose(hmd: ovrHmd; eye: ovrEyeType): ovrPosef; cdecl;



//------------------------------------------------------------------------------
// *****  Client Distortion Rendering Functions

// These functions provide the distortion data and render timing support necessary to allow
// client rendering of distortion. Client-side rendering involves the following steps:
//
//  1. Setup ovrEyeDesc based on the desired texture size and FOV.
//     Call ovrHmd_GetRenderDesc to get the necessary rendering parameters for each eye.
//
//  2. Use ovrHmd_CreateDistortionMesh to generate the distortion mesh.
//
//  3. Use ovrHmd_BeginFrameTiming, ovrHmd_GetEyePose, and ovrHmd_BeginFrameTiming
//     in the rendering loop to obtain timing and predicted head orientation when
//     rendering each eye.
//      - When using timewarp, use ovr_WaitTillTime after the rendering and gpu flush, followed
//        by ovrHmd_GetEyeTimewarpMatrices to obtain the timewarp matrices used
//        by the distortion pixel shader. This will minimize latency.
//

/// Computes the distortion viewport, view adjust, and other rendering parameters for
/// the specified eye. This can be used instead of ovrHmd_ConfigureRendering to do
/// setup for client rendered distortion.
function ovrHmd_GetRenderDesc(hmd: ovrHmd; eyeType: ovrEyeType;
  fov: ovrFovPort): ovrEyeRenderDesc; cdecl;


/// Describes a vertex used by the distortion mesh. This is intended to be converted into
/// the engine-specific format. Some fields may be unused based on the ovrDistortionCaps
/// flags selected. TexG and TexB, for example, are not used if chromatic correction is
/// not requested.
type
  ovrDistortionVertex = record
    ScreenPosNDC: ovrVector2f; // [-1,+1],[-1,+1] over the entire framebuffer.
    TimeWarpFactor: Single;    // Lerp factor between time-warp matrices. Can be encoded in Pos.z.
    VignetteFactor: Single;    // Vignette fade factor. Can be encoded in Pos.w.
    TanEyeAnglesR: ovrVector2f;
    TanEyeAnglesG: ovrVector2f;
    TanEyeAnglesB: ovrVector2f;
  end;
  ovrDistortionVertexArray = array[0..0] of ovrDistortionVertex;
  PovrDistortionVertexArray = ^ovrDistortionVertexArray;
  TWordArray = array[0..0] of Word;
  PWordArray = ^TWordArray;

/// Describes a full set of distortion mesh data, filled in by ovrHmd_CreateDistortionMesh.
/// Contents of this data structure, if not null, should be freed by ovrHmd_DestroyDistortionMesh.
  ovrDistortionMesh = record
    pVertexData: PovrDistortionVertexArray;
    pIndexData: PWordArray;
    VertexCount: Longword;
    IndexCount: Longword;
  end;
  PovrDistortionMesh = ^ovrDistortionMesh;

/// Generate distortion mesh per eye.
/// Distortion capabilities will depend on 'distortionCaps' flags. Users should
/// render using the appropriate shaders based on their settings.
/// Distortion mesh data will be allocated and written into the ovrDistortionMesh data structure,
/// which should be explicitly freed with ovrHmd_DestroyDistortionMesh.
/// Users should call ovrHmd_GetRenderScaleAndOffset to get uvScale and Offset values for rendering.
/// The function shouldn't fail unless theres is a configuration or memory error, in which case
/// ovrDistortionMesh values will be set to null.
/// This is the only function in the SDK reliant on eye relief, currently imported from profiles,
/// or overriden here.
function ovrHmd_CreateDistortionMesh(hmd: ovrHmd; eyeType: ovrEyeType;
  fov: ovrFovPort; distortionCaps: Longword;
  meshData: PovrDistortionMesh): ovrBool; cdecl;

/// Used to free the distortion mesh allocated by ovrHmd_GenerateDistortionMesh. meshData elements
/// are set to null and zeroes after the call.
procedure ovrHmd_DestroyDistortionMesh(meshData: PovrDistortionMesh); cdecl;

/// Computes updated 'uvScaleOffsetOut' to be used with a distortion if render target size or
/// viewport changes after the fact. This can be used to adjust render size every frame if desired.
type
  TuvScaleOffsetOut = array[0..1] of ovrVector2f;
procedure ovrHmd_GetRenderScaleAndOffset(fov: ovrFovPort; textureSize: ovrSizei;
  renderViewport: ovrRecti; uvScaleOffsetOut: TuvScaleOffsetOut); cdecl;


/// Thread-safe timing function for the main thread. Caller should increment frameIndex
/// with every frame and pass the index where applicable to functions called on the
/// rendering thread.
function ovrHmd_GetFrameTiming(hmd: ovrHmd; frameIndex: Longword): ovrFrameTiming; cdecl;

/// Called at the beginning of the frame on the rendering thread.
/// Pass frameIndex == 0 if ovrHmd_GetFrameTiming isn't being used. Otherwise,
/// pass the same frame index as was used for GetFrameTiming on the main thread.
function ovrHmd_BeginFrameTiming(hmd: ovrHmd; frameIndex: Longword): ovrFrameTiming; cdecl;

/// Marks the end of client distortion rendered frame, tracking the necessary timing information.
/// This function must be called immediately after Present/SwapBuffers + GPU sync. GPU sync is
/// important before this call to reduce latency and ensure proper timing.
procedure ovrHmd_EndFrameTiming(hmd: ovrHmd); cdecl;

/// Initializes and resets frame time tracking. This is typically not necessary, but
/// is helpful if game changes vsync state or video mode. vsync is assumed to be on if this
/// isn't called. Resets internal frame index to the specified number.
procedure ovrHmd_ResetFrameTiming(hmd: ovrHmd; frameIndex: Longword); cdecl;


/// Computes timewarp matrices used by distortion mesh shader, these are used to adjust
/// for head orientation change since the last call to ovrHmd_GetEyePose when rendering
/// this eye. The ovrDistortionVertex::TimeWarpFactor is used to blend between the
/// matrices, usually representing two different sides of the screen.
/// Must be called on the same thread as ovrHmd_BeginFrameTiming.
type
  TtwmOut = array[0..1] of ovrMatrix4f;
procedure ovrHmd_GetEyeTimewarpMatrices(hmd: ovrHmd; eye: ovrEyeType;
  renderPose: ovrPosef; twmOut: TtwmOut); cdecl;


//------------------------------------------------------------------------------
// ***** Stateless math setup functions

/// Used to generate projection from ovrEyeDesc::Fov.
function ovrMatrix4f_Projection(fov: ovrFovPort; znear, zfar: Single;
  rightHanded: ovrBool): ovrMatrix4f; cdecl;

/// Used for 2D rendering, Y is down
/// orthoScale = 1.0f / pixelsPerTanAngleAtCenter
/// orthoDistance = distance from camera, such as 0.8m
function ovrMatrix4f_OrthoSubProjection(projection: ovrMatrix4f;
  orthoScale: ovrVector2f; orthoDistance, eyeViewAdjustX: Single): ovrMatrix4f; cdecl;

/// Returns global, absolute high-resolution time in seconds. This is the same
/// value as used in sensor messages.
function ovr_GetTimeInSeconds: Double; cdecl;

/// Waits until the specified absolute time.
function ovr_WaitTillTime(absTime: Double): Double; cdecl;



// -----------------------------------------------------------------------------
// ***** Latency Test interface

/// Does latency test processing and returns 'TRUE' if specified rgb color should
/// be used to clear the screen.
type
  TrgbColorOut = array[0..2] of Byte;
function ovrHmd_ProcessLatencyTest(hmd: ovrHmd; rgbColorOut: TrgbColorOut): ovrBool; cdecl;

/// Returns non-null string once with latency test result, when it is available.
/// Buffer is valid until next call.
function ovrHmd_GetLatencyTestResult(hmd: ovrHmd): PAnsiChar; cdecl;

/// Returns the latency testing color in rgbColorOut to render when using a DK2
/// Returns false if this feature is disabled or not-applicable (e.g. using a DK1)
function ovrHmd_GetLatencyTest2DrawColor(hmd: ovrHmd; rgbColorOut: TrgbColorOut): ovrBool; cdecl;


//------------------------------------------------------------------------------
// ***** Health and Safety Warning Display interface
//

/// Used by ovrhmd_GetHSWDisplayState to report the current display state.
type
  ovrHSWDisplayState = record
    /// If true then the warning should be currently visible
    /// and the following variables have meaning. Else there is no
    /// warning being displayed for this application on the given HMD.
    Displayed: ovrBool;
    StartTime: Double;       /// Absolute time when the warning was first displayed. See ovr_GetTimeInSeconds().
    DismissibleTime: Double; /// Earliest absolute time when the warning can be dismissed. May be a time in the past.
  end;
  PovrHSWDisplayState = ^ovrHSWDisplayState;

/// Returns the current state of the HSW display. If the application is doing the rendering of
/// the HSW display then this function serves to indicate that the the warning should be
/// currently displayed. If the application is using SDK-based eye rendering then the SDK by
/// default automatically handles the drawing of the HSW display. An application that uses
/// application-based eye rendering should use this function to know when to start drawing the
/// HSW display itself and can optionally use it in conjunction with ovrhmd_DismissHSWDisplay
/// as described below.
///
/// Example usage for application-based rendering:
///    bool HSWDisplayCurrentlyDisplayed = false; // global or class member variable
///    ovrHSWDisplayState hswDisplayState;
///    ovrhmd_GetHSWDisplayState(Hmd, &hswDisplayState);
///
///    if (hswDisplayState.Displayed && !HSWDisplayCurrentlyDisplayed) {
///        <insert model into the scene that stays in front of the user>
///        HSWDisplayCurrentlyDisplayed = true;
///    }
procedure ovrHmd_GetHSWDisplayState(hmd: ovrHmd; hasWarningState: PovrHSWDisplayState); cdecl;

/// Dismisses the HSW display if the warning is dismissible and the earliest dismissal time
/// has occurred. Returns true if the display is valid and could be dismissed. The application
/// should recognize that the HSW display is being displayed (via ovrhmd_GetHSWDisplayState)
/// and if so then call this function when the appropriate user input to dismiss the warning
/// occurs.
///
/// Example usage :
///    void ProcessEvent(int key) {
///        if(key == escape) {
///            ovrHSWDisplayState hswDisplayState;
///            ovrhmd_GetHSWDisplayState(hmd, &hswDisplayState);
///
///            if(hswDisplayState.Displayed && ovrhmd_DismissHSWDisplay(hmd)) {
///                <remove model from the scene>
///                HSWDisplayCurrentlyDisplayed = false;
///            }
///        }
///    }
function ovrHmd_DismissHSWDisplay(hmd: ovrHmd): ovrBool; cdecl;





// -----------------------------------------------------------------------------
// ***** Property Access

// NOTICE: This is experimental part of API that is likely to go away or change.

// These allow accessing different properties of the HMD and profile.
// Some of the properties may go away with profile/HMD versions, so software should
// use defaults and/or proper fallbacks.
//

// For now, access profile entries; this will change.
{$IFNDEF OVR_KEY_USER}
const
  OVR_KEY_USER                 = 'User';
  OVR_KEY_NAME                 = 'Name';
  OVR_KEY_GENDER               = 'Gender';
  OVR_KEY_PLAYER_HEIGHT        = 'PlayerHeight';
  OVR_KEY_EYE_HEIGHT           = 'EyeHeight';
  OVR_KEY_IPD                  = 'IPD';
  OVR_KEY_NECK_TO_EYE_DISTANCE = 'NeckEyeDistance';
  OVR_KEY_CAMERA_POSITION      = 'CenteredFromWorld';

  // TODO: remove this duplication with OVR_Profile.h
  // Default measurements empirically determined at Oculus to make us happy
  // The neck model numbers were derived as an average of the male and female averages from ANSUR-88
  // NECK_TO_EYE_HORIZONTAL = H22 - H43 = INFRAORBITALE_BACK_OF_HEAD - TRAGION_BACK_OF_HEAD
  // NECK_TO_EYE_VERTICAL = H21 - H15 = GONION_TOP_OF_HEAD - ECTOORBITALE_TOP_OF_HEAD
  // These were determined to be the best in a small user study, clearly beating out the previous default values
  OVR_DEFAULT_GENDER                 = 'Unknown';
  OVR_DEFAULT_PLAYER_HEIGHT          = 1.778;
  OVR_DEFAULT_EYE_HEIGHT             = 1.675;
  OVR_DEFAULT_IPD                    = 0.064;
  OVR_DEFAULT_NECK_TO_EYE_HORIZONTAL = 0.0805;
  OVR_DEFAULT_NECK_TO_EYE_VERTICAL   = 0.075;
  OVR_DEFAULT_EYE_RELIEF_DIAL        = 3;
  OVR_DEFAULT_CAMERA_POSITION: array[0..6] of Integer = (0,0,0,1,0,0,0);
{$ENDIF}

/// Get boolean property. Returns first element if property is a boolean array.
/// Returns defaultValue if property doesn't exist.
function ovrHmd_GetBool(hmd: ovrHmd; propertyName: PAnsiChar; defaultVal: ovrBool): ovrBool; cdecl;

/// Modify bool property; false if property doesn't exist or is readonly.
function ovrHmd_SetBool(hmd: ovrHmd; propertyName: PAnsiChar; value: ovrBool): ovrBool; cdecl;

/// Get integer property. Returns first element if property is an integer array.
/// Returns defaultValue if property doesn't exist.
function ovrHmd_GetInt(hmd: ovrHmd; propertyName: PAnsiChar; defaultVal: Longint): Longint; cdecl;

/// Modify integer property; false if property doesn't exist or is readonly.
function ovrHmd_SetInt(hmd: ovrHmd; propertyName: PAnsiChar; value: Longint): ovrBool; cdecl;

/// Get float property. Returns first element if property is a float array.
/// Returns defaultValue if property doesn't exist.
function ovrHmd_GetFloat(hmd: ovrHmd; propertyName: PAnsiChar; defaultVal: Single): Single; cdecl;

/// Modify float property; false if property doesn't exist or is readonly.
function ovrHmd_SetFloat(hmd: ovrHmd; propertyName: PAnsiChar; value: Single): ovrBool; cdecl;

/// Get float[] property. Returns the number of elements filled in, 0 if property doesn't exist.
/// Maximum of arraySize elements will be written.
function ovrHmd_GetFloatArray(hmd: ovrHmd; propertyName: PAnsiChar;
  values: PSingle; arraySize: Longword): Longword; cdecl;

/// Modify float[] property; false if property doesn't exist or is readonly.
function ovrHmd_SetFloatArray(hmd: ovrHmd; propertyName: PAnsiChar;
  values: PSingle; arraySize: Longword): ovrBool; cdecl;

/// Get string property. Returns first element if property is a string array.
/// Returns defaultValue if property doesn't exist.
/// String memory is guaranteed to exist until next call to GetString or GetStringArray, or HMD is destroyed.
function ovrHmd_GetString(hmd: ovrHmd; propertyName, defaultVal: PAnsiChar): PAnsiChar; cdecl;

/// Set string property
function ovrHmd_SetString(hmd: ovrHmd; propertyName, value: PAnsiChar): ovrBool; cdecl;

implementation

const
  libovrdll = 'libovr.dll';

procedure ovr_InitializeRenderingShim; external libovrdll;
function ovr_Initialize; external libovrdll;
procedure ovr_Shutdown; external libovrdll;
function ovr_GetVersionString; external libovrdll;
function ovrHmd_Detect; external libovrdll;
function ovrHmd_Create; external libovrdll;
procedure ovrHmd_Destroy; external libovrdll;
function ovrHmd_CreateDebug; external libovrdll;
function ovrHmd_GetLastError; external libovrdll;
function ovrHmd_AttachToWindow(hmd: ovrHmd; window: THandle;
  const destMirrorRect: ovrRecti;
  const sourceRenderTargetRect: ovrRecti): ovrBool; cdecl; overload; external libovrdll;
function ovrHmd_AttachToWindow(hmd: ovrHmd; window: THandle;
  destMirrorRect: PovrRecti = nil;
  sourceRenderTargetRect: PovrRecti = nil): ovrBool; cdecl; overload; external libovrdll;
function ovrHmd_GetEnabledCaps; external libovrdll;
procedure ovrHmd_SetEnabledCaps; external libovrdll;
function ovrHmd_ConfigureTracking; external libovrdll;
procedure ovrHmd_RecenterPose; external libovrdll;
function ovrHmd_GetTrackingState; external libovrdll;
function _ovrHmd_GetFovTextureSize(hmd: ovrHmd; eye: ovrEyeType; fov: ovrFovPort;
  pixelsPerDisplayPixel: Single): PovrSizei; cdecl; external libovrdll name 'ovrHmd_GetFovTextureSize';
function ovrHmd_GetFovTextureSize(hmd: ovrHmd; eye: ovrEyeType; fov: ovrFovPort;
  pixelsPerDisplayPixel: Single): ovrSizei;
var
  tmp: ovrSizei;
begin
  _ovrHmd_GetFovTextureSize(hmd, eye, fov, pixelsPerDisplayPixel);
  asm
    MOV tmp.w,EAX
    MOV tmp.h,EDX
  end;
  Result := tmp;
end;
function ovrHmd_ConfigureRendering; external libovrdll;
function ovrHmd_BeginFrame; external libovrdll;
procedure ovrHmd_EndFrame; external libovrdll;
function ovrHmd_GetEyePose; external libovrdll;
function ovrHmd_GetRenderDesc; external libovrdll;
function ovrHmd_CreateDistortionMesh; external libovrdll;
procedure ovrHmd_DestroyDistortionMesh; external libovrdll;
procedure ovrHmd_GetRenderScaleAndOffset; external libovrdll;
function ovrHmd_GetFrameTiming; external libovrdll;
function ovrHmd_BeginFrameTiming; external libovrdll;
procedure ovrHmd_EndFrameTiming; external libovrdll;
procedure ovrHmd_ResetFrameTiming; external libovrdll;
procedure ovrHmd_GetEyeTimewarpMatrices; external libovrdll;
function ovrMatrix4f_Projection; external libovrdll;
function ovrMatrix4f_OrthoSubProjection; external libovrdll;
function ovr_GetTimeInSeconds; external libovrdll;
function ovr_WaitTillTime; external libovrdll;
function ovrHmd_ProcessLatencyTest; external libovrdll;
function ovrHmd_GetLatencyTestResult; external libovrdll;
function ovrHmd_GetLatencyTest2DrawColor; external libovrdll;
procedure ovrHmd_GetHSWDisplayState; external libovrdll;
function ovrHmd_DismissHSWDisplay; external libovrdll;
function ovrHmd_GetBool; external libovrdll;
function ovrHmd_SetBool; external libovrdll;
function ovrHmd_GetInt; external libovrdll;
function ovrHmd_SetInt; external libovrdll;
function ovrHmd_GetFloat; external libovrdll;
function ovrHmd_SetFloat; external libovrdll;
function ovrHmd_GetFloatArray; external libovrdll;
function ovrHmd_SetFloatArray; external libovrdll;
function ovrHmd_GetString; external libovrdll;
function ovrHmd_SetString; external libovrdll;

end.
