unit OpenGL.Context.Win;

interface
uses
  Winapi.Windows,
  OpenGL.Apis
  ;


function  InitOpenGL( TargetMajorVersion, TargetMinorVersion : integer): Boolean;
procedure CloseOpenGL;
function  InitOpenGLES( TargetMajorVersion, TargetMinorVersion : integer): Boolean;
procedure CloseOpenGLES;
function  CreateEGLRenderingContext(Wnd : EGLNativeWindowType; DC: EGLNativeDisplayType; Option : TGLOption; TargetMajorVersion, TargetMinorVersion : integer; SharedContext : EGLContext = nil): TGLContextInfo;
function  CreateWGLRenderingContext(DC: HDC; Option : TGLOption; TargetMajorVersion, TargetMinorVersion : integer; SharedContext : HGLRC = 0): TGLContextInfo;

implementation
uses
  System.SysUtils;

var
  EGLHandle : TGLLibHandle = 0;
  GLHandle : TGLLibHandle = 0;
  GLESHandle : TGLLibHandle = 0;
  FirstContext : Boolean = false;

function InitOpenGLESFromLibrary(EGL_Name, GLES_Name: String; TargetMajorVersion, TargetMinorVersion : integer): Boolean;

  function EGLGetProc(name : string; major, minor : integer; var validated_flag : Boolean) : Pointer;
  begin
    case major of
      PLATFORM_PROC: result := GetProcAddress(EGLHandle, PAnsiChar(TMarshal.AsAnsi(name)));
    else
      result := GetProcAddress(GLESHandle, PAnsiChar(TMarshal.AsAnsi(name)));
    end;

    if result = nil then validated_flag := false;
  end;

var
  OldError: Longint;
  fname : string;
begin
  Result := False;
  CloseOpenGLES;

  OldError := SetErrorMode(SEM_NOOPENFILEERRORBOX);
  try
    fname := ExtractFilePath(ParamStr(0))+EGL_Name;
    if FileEXists(fname) then
      EGLHandle := LoadLibrary(PChar(fname));

    fname := ExtractFilePath(ParamStr(0))+GLES_Name;
    if FileEXists(fname) then
      GLESHandle := LoadLibrary(PChar(fname));

    if (EGLHandle > 0) then
    begin
      LoadEGLCommands(@EGLGetProc);
      LoadOpenGLES_Commands(@EGLGetProc, TargetMajorVersion, TargetMinorVersion);
      FirstContext := true;
      Result:=True;
    end
    else
    begin
      if EGLHandle > 0 then FreeLibrary(EGLHandle);
      if GLESHandle > 0 then FreeLibrary(GLESHandle);
    end;
  finally
    SetErrorMode(OldError);
  end;
end;

function GetProc(name : string; major, minor : integer; var validated_flag : Boolean) : Pointer;
begin
  case major of
    PLATFORM_PROC: result := GetProcAddress(GLHandle, PAnsiChar(TMarshal.AsAnsi(name)));
    EXTENSION_PROC: result := wglGetProcAddress(PAnsiChar(TMarshal.AsAnsi(name)));
  else
    if (major = 1) and (minor = 0) then
      result := GetProcAddress(GLHandle, PAnsiChar(TMarshal.AsAnsi(name)))
    else
      result := wglGetProcAddress(PAnsiChar(TMarshal.AsAnsi(name)));
  end;
  if result = nil then validated_flag := false;
end;

function InitOpenGLFromLibrary(GL_Name: String; TargetMajorVersion, TargetMinorVersion : integer): Boolean;
var
  OldError: Longint;
begin
  Result := False;
  CloseOpenGL;

  OldError := SetErrorMode(SEM_NOOPENFILEERRORBOX);
  try
    GLHandle := LoadLibrary(PChar(GL_Name));

    if (GLHandle > 0) then
    begin
      LoadWGLCommands(@GetProc);
      LoadOpenGL_Commands(@GetProc, TargetMajorVersion, TargetMinorVersion);
      FirstContext := true;
      Result:=True;
    end;
  finally
    SetErrorMode(OldError);
  end;
end;

// refer : Mike Lischke's OpenGL.pas
function  CreateWGLRenderingContext(DC: HDC; Option : TGLOption; TargetMajorVersion, TargetMinorVersion : integer; SharedContext : HGLRC): TGLContextInfo;

  function CreateSharedContext(context : HGLRC) : HGLRC;
  var
     newcontext_attr : array[0..8] of GLint;
  begin
    if Assigned(wglCreateContextAttribsARB) then
    begin
      newcontext_attr[0] := WGL_CONTEXT_MAJOR_VERSION_ARB;
      newcontext_attr[1] := TargetMajorVersion;
      newcontext_attr[2] := WGL_CONTEXT_MINOR_VERSION_ARB;
      newcontext_attr[3] := TargetMinorVersion;
      newcontext_attr[4] := WGL_CONTEXT_FLAGS_ARB;
      newcontext_attr[5] := 0;
      newcontext_attr[6] := WGL_CONTEXT_PROFILE_MASK_ARB;
      newcontext_attr[7] := WGL_CONTEXT_CORE_PROFILE_BIT_ARB;
      newcontext_attr[8] := 0;

      result := wglCreateContextAttribsARB(DC, 0, @newcontext_attr[0]);
    end else Result := 0;
  end;

var
  PFDescriptor : TPixelFormatDescriptor;
  format : integer;
  temprc : HGLRC;
  exts : string;
begin
  FillChar(PFDescriptor,SizeOf(PFDescriptor),0);
  with PFDescriptor do
  begin
    nSize:=sizeof(PFDescriptor);
    nVersion:=1;
    dwFlags:=PFD_SUPPORT_OPENGL;
    if GetObjectType(DC) in [OBJ_MEMDC,OBJ_METADC,OBJ_ENHMETADC]
      then dwFlags:=dwFlags or PFD_DRAW_TO_BITMAP
      else dwFlags:=dwFlags or PFD_DRAW_TO_WINDOW;

    dwFlags:=dwFlags or Option.Flags;
    if Option.DoubleBuffered then dwFlags := dwFlags or PFD_DOUBLEBUFFER;

    iPixelType:=PFD_TYPE_RGBA;
    cColorBits := Option.RSize+Option.GSize+Option.BSize+Option.ASize;
    cDepthBits := Option.DepthSize;
    cStencilBits := Option.StencilSize;
    iLayerType:=Byte(PFD_MAIN_PLANE);
  end;

  InitOpenGL( TargetMajorVersion, TargetMinorVersion );

  format:= ChoosePixelFormat(DC, @PFDescriptor);
  if not Assigned(SetPixelFormat) then
  begin
    SetPixelFormat := @Winapi.Windows.SetPixelFormat;
  end;
  SetPixelFormat(DC, format, @PFDescriptor);

  DescribePixelFormat(DC, format, SizeOf(PFDescriptor), @PFDescriptor);

  result.DC := DC;

  if FirstContext then
  begin
    Result.RenderingContext := wglCreateContext(DC);

    if Result.RenderingContext > 0 then
    begin
      wglMakeCurrent(DC, Result.RenderingContext);

      exts := string(glGetString(GL_EXTENSIONS))+' ';
      {$if Declared (wglGetExtensionsStringARB)}
      wglGetExtensionsStringARB := wglGetProcAddress('wglGetExtensionsStringARB');
      if Assigned(wglGetExtensionsStringARB) then
      begin
        exts := exts + string(wglGetExtensionsStringARB(DC)) + ' ';
      end;
      {$endif}

      {$if Declared (wglGetExtensionsStringEXT)}
      wglGetExtensionsStringEXT := wglGetProcAddress('wglGetExtensionsStringEXT');
      if Assigned(wglGetExtensionsStringEXT) then
      begin
        exts := exts + string(wglGetExtensionsStringEXT()) + ' ';
      end;
      {$endif}

      LoadWGLExtensionCommands(@GetProc, exts);

      if Assigned(wglCreateContextAttribsARB) and (TargetMajorVersion >= 3) then
      begin
        temprc := result.RenderingContext;
        result.RenderingContext := CreateSharedContext(0);
        wglMakeCurrent(DC, 0);
        wglDeleteContext(temprc);
        wglMakeCurrent(DC, Result.RenderingContext);
        LoadOpenGL_Commands(@GetProc, TargetMajorVersion, TargetMinorVersion);
      end;

      LoadOpenGL_ExtensionCommands(@GetProc,exts, TargetMajorVersion, TargetMinorVersion);
      FirstContext:=False;
      OpenGLVersionString := string(glGetString(GL_VERSION));
      wglMakeCurrent(0,0);
    end;
  end else begin
    if Assigned(wglCreateContextAttribsARB) and (SharedContext <> 0) then
    begin
      result.RenderingContext := CreateSharedContext(SharedContext);
    end else begin
      result.RenderingContext := wglCreateContext(DC);
    end;
    wglMakeCurrent(0,0);
  end;
end;

procedure CloseOpenGLES;
begin
  if EGLHandle > 0 then begin FreeLibrary(EGLHandle); EGLHandle:=0; end;
  if GLESHandle > 0 then begin FreeLibrary(GLESHandle); GLESHandle:=0; end;
  ClearCommands;
  EGLMode := false;
end;

function InitOpenGL( TargetMajorVersion, TargetMinorVersion : integer): Boolean;
begin
  if (GLHandle = 0) then
  begin
    Result := InitOpenGLFromLibrary('OpenGL32.DLL', TargetMajorVersion, TargetMinorVersion );
  end
  else Result := True;
  EGLMode := false;
end;

procedure CloseOpenGL;
begin
  if GLHandle > 0 then begin FreeLibrary(GLHandle); GLHandle:=0; end;
  ClearCommands;
end;

function InitOpenGLES( TargetMajorVersion, TargetMinorVersion : integer): Boolean;
begin
  if (EGLHandle = 0) then
  begin
    Result := InitOpenGLESFromLibrary('libegl.dll','libglesv2.dll', TargetMajorVersion, TargetMinorVersion);
  end
  else Result := True;
  EGLMode := true;
end;

function  CreateEGLRenderingContext(Wnd : EGLNativeWindowType; DC: EGLNativeDisplayType;  Option : TGLOption; TargetMajorVersion, TargetMinorVersion : integer; SharedContext : EGLContext): TGLContextInfo;

  function EGLGetProc_ext(name : string; var validated_flag : Boolean) : Pointer;
  begin
    if not validated_flag then
    begin
      result := nil;
      exit;
    end;
    result := eglGetProcAddress(PAnsiChar(TMarshal.AsAnsi(name)));
    if result = nil then validated_flag := false;
  end;

const
  ContextAttributes: array[0..2] of EGLint = (EGL_CONTEXT_CLIENT_VERSION, 2, EGL_NONE);
var
  config : EGLConfig;
  attrib : array[0..14] of EGLint;
  config_cnt : EGLint;
  format : EGLint;
  eglMajorVer, eglMinorVer : EGLint;
  exts : string;
begin
  InitOpenGLES( TargetMajorVersion, TargetMinorVersion );

  result.Display := eglGetDisplay(DC);
  if eglInitialize(result.Display, @eglMajorVer, @eglMinorVer) <> EGL_FALSE then
  begin
//    eglBindAPI(EGL_OPENGL_ES_API);
    attrib[0] := EGL_RED_SIZE;
    attrib[1] := Option.RSize;
    attrib[2] := EGL_GREEN_SIZE;
    attrib[3] := Option.GSize;
    attrib[4] := EGL_BLUE_SIZE;
    attrib[5] := Option.BSize;
    attrib[6] := EGL_ALPHA_SIZE;
    attrib[7] := Option.ASize;
    attrib[8] := EGL_DEPTH_SIZE;
    attrib[9] := Option.DepthSize;
    attrib[10] := EGL_SAMPLE_BUFFERS;
    attrib[11] := 0;//Ord(Option.DoubleBuffered);
    attrib[12] := EGL_STENCIL_SIZE;
    attrib[13] := Option.StencilSize;
    attrib[14] := EGL_NONE;
    if eglChooseConfig(result.Display, @attrib[0], @config, 1, @config_cnt) <> EGL_FALSE then
    begin
      eglGetConfigAttrib(result.Display, @Config, EGL_NATIVE_VISUAL_ID, @format);
      result.Surface := eglCreateWindowSurface(result.Display, config, Wnd, nil);
      result.Context := eglCreateContext(result.Display, config, SharedContext, @ContextAttributes[0]);

  // read implementation properties
      if FirstContext and (result.Context <> nil) then
      begin
        if eglMakeCurrent(result.Display, result.Surface, result.Surface, result.Context) <> EGL_FALSE then
        begin
          exts := string(eglQueryString(result.Display, GL_EXTENSIONS)) + ' ';
          LoadEGLExtensionCommands(result.Display, @EGLGetProc_ext, exts);
          LoadOpenGLES_ExtensionCommands(@EGLGetProc_ext, exts, TargetMajorVersion, TargetMinorVersion);
          OpenGLVersionString := string(glGetString(GL_VERSION));
          eglMakeCurrent(result.Display, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);
        end;
        FirstContext := false;
      end;
    end;
  end;
end;

end.
